@@ -26,6 +26,13 @@
  *
  *  ATTENTION!   ATTENTION!   ATTENTION!   ATTENTION!   ATTENTION! */
 
+/*  Dec 20, 2002
+ *
+ *  Initial test implementation of strcoll, strxfrm, wcscoll, and wcsxfrm.
+ *  The code needs to be cleaned up a good bit, but I'd like to see people
+ *  test it out.
+ */
+
 #define _STDIO_UTILITY
 #define _GNU_SOURCE
 #include <string.h>
@@ -36,11 +43,12 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <signal.h>
+#include <assert.h>
+#include <locale.h>
 
 #ifdef WANT_WIDE
 #include <wchar.h>
 #include <wctype.h>
-#include <locale.h>
 
 #define Wvoid			wchar_t
 #define Wchar			wchar_t
@@ -62,6 +70,13 @@ typedef unsigned char	__string_uchar_t;
  * have to be changed! */
 
 #define _SYS_NERR			125
+#if defined(__mips__) || defined(__sparc__)
+/* sparce and mips have an extra error entry, as EDEADLK and EDEADLOCK have
+ * different meanings on those platforms. */
+#undef _SYS_NERR
+#define _SYS_NERR			126
+#endif
+
 #define _SYS_ERRMSG_MAXLEN	 50
 
 extern const char _string_syserrmsgs[];
@@ -213,6 +228,12 @@ const char _string_syserrmsgs[] = {
 	/* 122: 2852, 20 */ "Disk quota exceeded\0"
 	/* 123: 2872, 16 */ "No medium found\0"
 	/* 124: 2888, 18 */ "Wrong medium type"
+#if defined(__mips__) || defined(__sparc__)
+	"\0"
+	/* 125: 2906, 28 */ "File locking deadlock error"
+#endif
+	/* Note: for mips we are ignoring ECANCELED since glibc doesn't have a
+	 * corresponsding message.*/
 };
 
 #endif
@@ -221,138 +242,160 @@ const char _string_syserrmsgs[] = {
 
 link_warning(_sys_errlist, "sys_nerr and sys_errlist are obsolete and uClibc support for them (in at least some configurations) will probably be unavailable in the near future.")
 
-int sys_nerr = _SYS_NERR;
-
 const char *const sys_errlist[] = {
-	_string_syserrmsgs + 0,
-	_string_syserrmsgs + 8,
-	_string_syserrmsgs + 32,
-	_string_syserrmsgs + 58,
-	_string_syserrmsgs + 74,
-	_string_syserrmsgs + 98,
-	_string_syserrmsgs + 117,
-	_string_syserrmsgs + 143,
-	_string_syserrmsgs + 166,
-	_string_syserrmsgs + 184,
-	_string_syserrmsgs + 204,
-	_string_syserrmsgs + 223,
-	_string_syserrmsgs + 256,
-	_string_syserrmsgs + 279,
-	_string_syserrmsgs + 297,
-	_string_syserrmsgs + 309,
-	_string_syserrmsgs + 331,
-	_string_syserrmsgs + 355,
-	_string_syserrmsgs + 367,
-	_string_syserrmsgs + 393,
-	_string_syserrmsgs + 408,
-	_string_syserrmsgs + 424,
-	_string_syserrmsgs + 439,
-	_string_syserrmsgs + 456,
-	_string_syserrmsgs + 486,
-	_string_syserrmsgs + 506,
-	_string_syserrmsgs + 537,
-	_string_syserrmsgs + 552,
-	_string_syserrmsgs + 567,
-	_string_syserrmsgs + 591,
-	_string_syserrmsgs + 604,
-	_string_syserrmsgs + 626,
-	_string_syserrmsgs + 641,
-	_string_syserrmsgs + 653,
-	_string_syserrmsgs + 686,
-	_string_syserrmsgs + 716,
-	_string_syserrmsgs + 742,
-	_string_syserrmsgs + 761,
-	_string_syserrmsgs + 780,
-	_string_syserrmsgs + 805,
-	_string_syserrmsgs + 825,
+	[0] =				_string_syserrmsgs + 0,
+	[EPERM] =			_string_syserrmsgs + 8,
+	[ENOENT] =			_string_syserrmsgs + 32,
+	[ESRCH] =			_string_syserrmsgs + 58,
+	[EINTR] =			_string_syserrmsgs + 74,
+	[EIO] =				_string_syserrmsgs + 98,
+	[ENXIO] =			_string_syserrmsgs + 117,
+	[E2BIG] =			_string_syserrmsgs + 143,
+	[ENOEXEC] =			_string_syserrmsgs + 166,
+	[EBADF] =			_string_syserrmsgs + 184,
+	[ECHILD] =			_string_syserrmsgs + 204,
+	[EAGAIN] =			_string_syserrmsgs + 223,
+	[ENOMEM] =			_string_syserrmsgs + 256,
+	[EACCES] =			_string_syserrmsgs + 279,
+	[EFAULT] =			_string_syserrmsgs + 297,
+	[ENOTBLK] =			_string_syserrmsgs + 309,
+	[EBUSY] =			_string_syserrmsgs + 331,
+	[EEXIST] =			_string_syserrmsgs + 355,
+	[EXDEV] =			_string_syserrmsgs + 367,
+	[ENODEV] =			_string_syserrmsgs + 393,
+	[ENOTDIR] =			_string_syserrmsgs + 408,
+	[EISDIR] =			_string_syserrmsgs + 424,
+	[EINVAL] =			_string_syserrmsgs + 439,
+	[ENFILE] =			_string_syserrmsgs + 456,
+	[EMFILE] =			_string_syserrmsgs + 486,
+	[ENOTTY] =			_string_syserrmsgs + 506,
+	[ETXTBSY] =			_string_syserrmsgs + 537,
+	[EFBIG] =			_string_syserrmsgs + 552,
+	[ENOSPC] =			_string_syserrmsgs + 567,
+	[ESPIPE] =			_string_syserrmsgs + 591,
+	[EROFS] =			_string_syserrmsgs + 604,
+	[EMLINK] =			_string_syserrmsgs + 626,
+	[EPIPE] =			_string_syserrmsgs + 641,
+	[EDOM] =			_string_syserrmsgs + 653,
+	[ERANGE] =			_string_syserrmsgs + 686,
+	[EDEADLK] =			_string_syserrmsgs + 716,
+	[ENAMETOOLONG] =	_string_syserrmsgs + 742,
+	[ENOLCK] =			_string_syserrmsgs + 761,
+	[ENOSYS] =			_string_syserrmsgs + 780,
+	[ENOTEMPTY] =		_string_syserrmsgs + 805,
+	[ELOOP] =			_string_syserrmsgs + 825,
 	/*  	_string_syserrmsgs + 859, */
-	NULL,						/* glibc compatiblity :-( */
-	_string_syserrmsgs + 860,
-	_string_syserrmsgs + 887,
-	_string_syserrmsgs + 906,
-	_string_syserrmsgs + 934,
-	_string_syserrmsgs + 959,
-	_string_syserrmsgs + 974,
-	_string_syserrmsgs + 988,
-	_string_syserrmsgs + 1013,
-	_string_syserrmsgs + 1042,
-	_string_syserrmsgs + 1069,
-	_string_syserrmsgs + 1084,
-	_string_syserrmsgs + 1101,
-	_string_syserrmsgs + 1128,
-	_string_syserrmsgs + 1142,
-	_string_syserrmsgs + 1151,
-	_string_syserrmsgs + 1172,
+	[ENOMSG] =			_string_syserrmsgs + 860,
+	[EIDRM] =			_string_syserrmsgs + 887,
+	[ECHRNG] =			_string_syserrmsgs + 906,
+	[EL2NSYNC] =		_string_syserrmsgs + 934,
+	[EL3HLT] =			_string_syserrmsgs + 959,
+	[EL3RST] =			_string_syserrmsgs + 974,
+	[ELNRNG] =			_string_syserrmsgs + 988,
+	[EUNATCH] =			_string_syserrmsgs + 1013,
+	[ENOCSI] =			_string_syserrmsgs + 1042,
+	[EL2HLT] =			_string_syserrmsgs + 1069,
+	[EBADE] =			_string_syserrmsgs + 1084,
+	[EBADR] =			_string_syserrmsgs + 1101,
+	[EXFULL] =			_string_syserrmsgs + 1128,
+	[ENOANO] =			_string_syserrmsgs + 1142,
+	[EBADRQC] =			_string_syserrmsgs + 1151,
+	[EBADSLT] =			_string_syserrmsgs + 1172,
 	/*  	_string_syserrmsgs + 1185, */
-	NULL,						/* glibc compatiblity :-( */
-	_string_syserrmsgs + 1186,
-	_string_syserrmsgs + 1207,
-	_string_syserrmsgs + 1227,
-	_string_syserrmsgs + 1245,
-	_string_syserrmsgs + 1259,
-	_string_syserrmsgs + 1284,
-	_string_syserrmsgs + 1314,
-	_string_syserrmsgs + 1336,
-	_string_syserrmsgs + 1353,
-	_string_syserrmsgs + 1375,
-	_string_syserrmsgs + 1391,
-	_string_syserrmsgs + 1405,
-	_string_syserrmsgs + 1433,
-	_string_syserrmsgs + 1448,
-	_string_syserrmsgs + 1467,
-	_string_syserrmsgs + 1486,
-	_string_syserrmsgs + 1498,
-	_string_syserrmsgs + 1536,
-	_string_syserrmsgs + 1563,
-	_string_syserrmsgs + 1592,
-	_string_syserrmsgs + 1615,
-	_string_syserrmsgs + 1654,
-	_string_syserrmsgs + 1691,
-	_string_syserrmsgs + 1723,
-	_string_syserrmsgs + 1771,
-	_string_syserrmsgs + 1809,
-	_string_syserrmsgs + 1859,
-	_string_syserrmsgs + 1903,
-	_string_syserrmsgs + 1922,
-	_string_syserrmsgs + 1937,
-	_string_syserrmsgs + 1968,
-	_string_syserrmsgs + 1997,
-	_string_syserrmsgs + 2014,
-	_string_syserrmsgs + 2045,
-	_string_syserrmsgs + 2068,
-	_string_syserrmsgs + 2091,
-	_string_syserrmsgs + 2117,
-	_string_syserrmsgs + 2141,
-	_string_syserrmsgs + 2171,
-	_string_syserrmsgs + 2212,
-	_string_syserrmsgs + 2235,
-	_string_syserrmsgs + 2267,
-	_string_syserrmsgs + 2283,
-	_string_syserrmsgs + 2306,
-	_string_syserrmsgs + 2342,
-	_string_syserrmsgs + 2375,
-	_string_syserrmsgs + 2400,
-	_string_syserrmsgs + 2426,
-	_string_syserrmsgs + 2466,
-	_string_syserrmsgs + 2502,
-	_string_syserrmsgs + 2548,
-	_string_syserrmsgs + 2583,
-	_string_syserrmsgs + 2604,
-	_string_syserrmsgs + 2623,
-	_string_syserrmsgs + 2636,
-	_string_syserrmsgs + 2653,
-	_string_syserrmsgs + 2683,
-	_string_syserrmsgs + 2709,
-	_string_syserrmsgs + 2731,
-	_string_syserrmsgs + 2756,
-	_string_syserrmsgs + 2784,
-	_string_syserrmsgs + 2814,
-	_string_syserrmsgs + 2835,
-	_string_syserrmsgs + 2852,
-	_string_syserrmsgs + 2872,
-	_string_syserrmsgs + 2888,
+	[EBFONT] =			_string_syserrmsgs + 1186,
+	[ENOSTR] =			_string_syserrmsgs + 1207,
+	[ENODATA] =			_string_syserrmsgs + 1227,
+	[ETIME] =			_string_syserrmsgs + 1245,
+	[ENOSR] =			_string_syserrmsgs + 1259,
+	[ENONET] =			_string_syserrmsgs + 1284,
+	[ENOPKG] =			_string_syserrmsgs + 1314,
+	[EREMOTE] =			_string_syserrmsgs + 1336,
+	[ENOLINK] =			_string_syserrmsgs + 1353,
+	[EADV] =			_string_syserrmsgs + 1375,
+	[ESRMNT] =			_string_syserrmsgs + 1391,
+	[ECOMM] =			_string_syserrmsgs + 1405,
+	[EPROTO] =			_string_syserrmsgs + 1433,
+	[EMULTIHOP] =		_string_syserrmsgs + 1448,
+	[EDOTDOT] =			_string_syserrmsgs + 1467,
+	[EBADMSG] =			_string_syserrmsgs + 1486,
+	[EOVERFLOW] =		_string_syserrmsgs + 1498,
+	[ENOTUNIQ] =		_string_syserrmsgs + 1536,
+	[EBADFD] =			_string_syserrmsgs + 1563,
+	[EREMCHG] =			_string_syserrmsgs + 1592,
+	[ELIBACC] =			_string_syserrmsgs + 1615,
+	[ELIBBAD] =			_string_syserrmsgs + 1654,
+	[ELIBSCN] =			_string_syserrmsgs + 1691,
+	[ELIBMAX] =			_string_syserrmsgs + 1723,
+	[ELIBEXEC] =		_string_syserrmsgs + 1771,
+	[EILSEQ] =			_string_syserrmsgs + 1809,
+	[ERESTART] =		_string_syserrmsgs + 1859,
+	[ESTRPIPE] =		_string_syserrmsgs + 1903,
+	[EUSERS] =			_string_syserrmsgs + 1922,
+	[ENOTSOCK] =		_string_syserrmsgs + 1937,
+	[EDESTADDRREQ] =	_string_syserrmsgs + 1968,
+	[EMSGSIZE] =		_string_syserrmsgs + 1997,
+	[EPROTOTYPE] =		_string_syserrmsgs + 2014,
+	[ENOPROTOOPT] =		_string_syserrmsgs + 2045,
+	[EPROTONOSUPPORT] =	_string_syserrmsgs + 2068,
+	[ESOCKTNOSUPPORT] =	_string_syserrmsgs + 2091,
+	[EOPNOTSUPP] =		_string_syserrmsgs + 2117,
+	[EPFNOSUPPORT] =	_string_syserrmsgs + 2141,
+	[EAFNOSUPPORT] =	_string_syserrmsgs + 2171,
+	[EADDRINUSE] =		_string_syserrmsgs + 2212,
+	[EADDRNOTAVAIL] =	_string_syserrmsgs + 2235,
+	[ENETDOWN] =		_string_syserrmsgs + 2267,
+	[ENETUNREACH] =		_string_syserrmsgs + 2283,
+	[ENETRESET] =		_string_syserrmsgs + 2306,
+	[ECONNABORTED] =	_string_syserrmsgs + 2342,
+	[ECONNRESET] =		_string_syserrmsgs + 2375,
+	[ENOBUFS] =			_string_syserrmsgs + 2400,
+	[EISCONN] =			_string_syserrmsgs + 2426,
+	[ENOTCONN] =		_string_syserrmsgs + 2466,
+	[ESHUTDOWN] =		_string_syserrmsgs + 2502,
+	[ETOOMANYREFS] =	_string_syserrmsgs + 2548,
+	[ETIMEDOUT] =		_string_syserrmsgs + 2583,
+	[ECONNREFUSED] =	_string_syserrmsgs + 2604,
+	[EHOSTDOWN] =		_string_syserrmsgs + 2623,
+	[EHOSTUNREACH] =	_string_syserrmsgs + 2636,
+	[EALREADY] =		_string_syserrmsgs + 2653,
+	[EINPROGRESS] =		_string_syserrmsgs + 2683,
+	[ESTALE] =			_string_syserrmsgs + 2709,
+	[EUCLEAN] =			_string_syserrmsgs + 2731,
+	[ENOTNAM] =			_string_syserrmsgs + 2756,
+	[ENAVAIL] =			_string_syserrmsgs + 2784,
+	[EISNAM] =			_string_syserrmsgs + 2814,
+	[EREMOTEIO] =		_string_syserrmsgs + 2835,
+	[EDQUOT] =			_string_syserrmsgs + 2852,
+	[ENOMEDIUM] =		_string_syserrmsgs + 2872,
+	[EMEDIUMTYPE] =		_string_syserrmsgs + 2888,
+
+#if EDEADLOCK != EDEADLK
+	[EDEADLOCK] =		_string_syserrmsgs + 2906,
+#endif
+
+#if EWOULDBLOCK != EAGAIN
+#error EWOULDBLOCK does not equal EAGAIN
+#endif
+
+	/* For now, ignore the other arch-specific errors.  glibc only maps EPROCLIM. */
+
+	/* some other mips errors */
+#ifdef ECANCELED
+#endif
+#ifdef EINIT
+#endif
+#ifdef EREMDEV
+#endif
+
+	/* some other sparc errors */
+#ifdef EPROCLIM
+#endif
+#ifdef ERREMOTE
+#endif
 };
 
+int sys_nerr = sizeof(sys_errlist)/sizeof(sys_errlist[0]);
+
 #endif
 /**********************************************************************/
 #ifdef L_wmemcpy
@@ -592,10 +635,13 @@ int Wmemcmp(const Wvoid *s1, const Wvoid *s2, size_t n)
 
 #ifdef L_strcmp
 
-#ifndef L_wcscmp
-#warning implement strcoll and remove weak alias (or enable for C locale only)
+#ifdef __LOCALE_C_ONLY
+#ifdef L_wcscmp
+weak_alias(wcscmp,wcscoll);
+#else  /* L_wcscmp */
 weak_alias(strcmp,strcoll);
-#endif
+#endif /* L_wcscmp */
+#endif /* __LOCALE_C_ONLY */
 
 int Wstrcmp(register const Wchar *s1, register const Wchar *s2)
 {
@@ -617,23 +663,6 @@ int Wstrcmp(register const Wchar *s1, register const Wchar *s2)
 	return r;
 #endif
 }
-#endif
-/**********************************************************************/
-#ifdef L_strcoll
-#error implement strcoll and remove weak_alias!!
-
-#if 0
-extern unsigned char *_ctype_collate;
-int strcoll(register const char *s1, const char *s2)
-{
-    int r;
-
-    while (!(r = (_ctype_collate[(int)(*s1++)]-_ctype_collate[(int)(*s2++)])));
-
-    return r;
-}
-#endif
-
 #endif
 /**********************************************************************/
 #ifdef L_wcsncmp
@@ -671,11 +700,6 @@ int Wstrncmp(register const Wchar *s1, register const Wchar *s2, size_t n)
 
 #endif
 /**********************************************************************/
-#ifdef L_strxfrm
-#error implement strxfrm
-/* size_t strxfrm(char *dst, const char *src, size_t len); */
-#endif	
-/**********************************************************************/
 #ifdef L_wmemchr
 #define L_memchr
 #define Wmemchr wmemchr
@@ -876,6 +900,9 @@ Wchar *Wstrstr(const Wchar *s1, const Wchar *s2)
 
 #endif
 /**********************************************************************/
+#undef Wstrspn
+#undef Wstrpbrk
+
 #ifdef L_wcstok
 #define L_strtok_r
 #define Wstrtok_r wcstok
@@ -1160,6 +1187,9 @@ void *memccpy(void * __restrict s1, const void * __restrict s2, int c, size_t n)
 
 #endif
 /**********************************************************************/
+#undef Wstrlen
+#undef Wstrcpy
+
 #ifdef L_wcsdup
 #define L_strdup
 #define Wstrdup wcsdup
@@ -1192,7 +1222,9 @@ char *strerror(int errnum)
 {
     static char buf[_SYS_ERRMSG_MAXLEN];
 
-    return (_susv3_strerror_r(errnum, buf, sizeof(buf)) == 0) ? buf : NULL;
+	_susv3_strerror_r(errnum, buf, sizeof(buf));
+
+	return buf;
 }
 
 #endif
@@ -1201,6 +1233,146 @@ char *strerror(int errnum)
 /**********************************************************************/
 #ifdef L__susv3_strerror_r
 
+#if defined(__alpha__) || defined(__mips__) || defined(__sparc__)
+
+static const unsigned char estridx[] = {
+	0,							/* success is always 0 */
+	EPERM,
+	ENOENT,
+	ESRCH,
+	EINTR,
+	EIO,
+	ENXIO,
+	E2BIG,
+	ENOEXEC,
+	EBADF,
+	ECHILD,
+	EAGAIN,
+	ENOMEM,
+	EACCES,
+	EFAULT,
+	ENOTBLK,
+	EBUSY,
+	EEXIST,
+	EXDEV,
+	ENODEV,
+	ENOTDIR,
+	EISDIR,
+	EINVAL,
+	ENFILE,
+	EMFILE,
+	ENOTTY,
+	ETXTBSY,
+	EFBIG,
+	ENOSPC,
+	ESPIPE,
+	EROFS,
+	EMLINK,
+	EPIPE,
+	EDOM,
+	ERANGE,
+	EDEADLK,
+	ENAMETOOLONG,
+	ENOLCK,
+	ENOSYS,
+	ENOTEMPTY,
+	ELOOP,
+	0,
+	ENOMSG,
+	EIDRM,
+	ECHRNG,
+	EL2NSYNC,
+	EL3HLT,
+	EL3RST,
+	ELNRNG,
+	EUNATCH,
+	ENOCSI,
+	EL2HLT,
+	EBADE,
+	EBADR,
+	EXFULL,
+	ENOANO,
+	EBADRQC,
+	EBADSLT,
+	0,
+	EBFONT,
+	ENOSTR,
+	ENODATA,
+	ETIME,
+	ENOSR,
+	ENONET,
+	ENOPKG,
+	EREMOTE,
+	ENOLINK,
+	EADV,
+	ESRMNT,
+	ECOMM,
+	EPROTO,
+	EMULTIHOP,
+	EDOTDOT,
+	EBADMSG,
+	EOVERFLOW,
+	ENOTUNIQ,
+	EBADFD,
+	EREMCHG,
+	ELIBACC,
+	ELIBBAD,
+	ELIBSCN,
+	ELIBMAX,
+	ELIBEXEC,
+	EILSEQ,
+	ERESTART,
+	ESTRPIPE,
+	EUSERS,
+	ENOTSOCK,
+	EDESTADDRREQ,
+	EMSGSIZE,
+	EPROTOTYPE,
+	ENOPROTOOPT,
+	EPROTONOSUPPORT,
+	ESOCKTNOSUPPORT,
+	EOPNOTSUPP,
+	EPFNOSUPPORT,
+	EAFNOSUPPORT,
+	EADDRINUSE,
+	EADDRNOTAVAIL,
+	ENETDOWN,
+	ENETUNREACH,
+	ENETRESET,
+	ECONNABORTED,
+	ECONNRESET,
+	ENOBUFS,
+	EISCONN,
+	ENOTCONN,
+	ESHUTDOWN,
+	ETOOMANYREFS,
+	ETIMEDOUT,
+	ECONNREFUSED,
+	EHOSTDOWN,
+	EHOSTUNREACH,
+	EALREADY,
+	EINPROGRESS,
+	ESTALE,
+	EUCLEAN,
+	ENOTNAM,
+	ENAVAIL,
+	EISNAM,
+	EREMOTEIO,
+#ifdef __mips__
+	0,							/* mips has an outrageous value for this... */
+#else
+	EDQUOT,
+#endif
+	ENOMEDIUM,
+	EMEDIUMTYPE,
+#if defined(__mips__) || defined(__sparc__)
+	EDEADLOCK,
+#endif
+};
+
+#endif
+
+
 int _susv3_strerror_r(int errnum, char *strerrbuf, size_t buflen)
 {
     register char *s;
@@ -1212,12 +1384,32 @@ int _susv3_strerror_r(int errnum, char *strerrbuf, size_t buflen)
 
     retval = EINVAL;
 
-    if (((unsigned int) errnum) < _SYS_NERR) {
+
+#if defined(__alpha__) || defined(__mips__) || defined(__sparc__)
+	/* Need to translate errno to string index. */
+	for (i = 0 ; i < sizeof(estridx)/sizeof(estridx[0]) ; i++) {
+		if (estridx[i] == errnum) {
+			goto GOT_ESTRIDX;
+		}
+	}
+	i = INT_MAX;	/* Failed, but may need to check mips special case. */
+#ifdef __mips__
+	if (errnum == EDQUOT) {	/* Deal with large EDQUOT value on mips */
+		i = 122;
+	}
+#endif /* __mips__ */
+ GOT_ESTRIDX:
+#else
+	/* No errno to string index translation needed. */
+	i = errnum;
+#endif
+
+    if (((unsigned int) i) < _SYS_NERR) {
 		/* Trade time for space.  This function should rarely be called
 		 * so rather than keeping an array of pointers for the different
 		 * messages, just run through the buffer until we find the
 		 * correct string. */
-		for (s = (char *) _string_syserrmsgs, i = errnum ; i ; ++s) {
+		for (s = (char *) _string_syserrmsgs ; i ; ++s) {
 			if (!*s) {
 				--i;
 			}
@@ -1268,6 +1460,40 @@ char *_glibc_strerror_r(int errnum, char *strerrbuf, size_t buflen)
     return strerrbuf;
 }
 
+#endif
+/**********************************************************************/
+#ifdef L_memmem
+
+void *memmem(const void *haystack, size_t haystacklen,
+		     const void *needle, size_t needlelen)
+{
+	register const char *ph;
+	register const char *pn;
+	const char *plast;
+	size_t n;
+
+	if (needlelen == 0) {
+		return (void *) haystack;
+	}
+
+	if (haystacklen >= needlelen) {
+		ph = (const char *) haystack;
+		pn = (const char *) needle;
+		plast = ph + (haystacklen - needlelen);
+
+		do {
+			n = 0;
+			while (ph[n] == pn[n]) {
+				if (++n == needlelen) {
+					return (void *) ph;
+				}
+			}
+		} while (++ph <= plast);
+	}
+
+	return NULL;
+}
+
 #endif
 /**********************************************************************/
 #ifdef L_wmempcpy
@@ -1547,14 +1773,19 @@ char *strsep(char ** __restrict s1, const char * __restrict s2)
 /**********************************************************************/
 #ifdef L_wcschrnul
 #define L_strchrnul
+#define __Wstrchrnul __wcschrnul
 #define Wstrchrnul wcschrnul
 #else
+#define __Wstrchrnul __strchrnul
 #define Wstrchrnul strchrnul
 #endif
 
 #ifdef L_strchrnul
 
-Wchar *Wstrchrnul(register const Wchar *s, Wint c)
+extern Wchar *__Wstrchrnul(register const Wchar *s, Wint c);
+weak_alias(__Wstrchrnul, Wstrchrnul);
+
+Wchar *__Wstrchrnul(register const Wchar *s, Wint c)
 {
 	--s;
 	while (*++s && (*s != ((Wchar)c)));
@@ -1708,18 +1939,40 @@ size_t strlcat(register char *__restrict dst,
 
 #endif
 /**********************************************************************/
+#ifdef WANT_WIDE
+extern size_t __wcslcpy(wchar_t *__restrict dst,
+						const wchar_t *__restrict src,
+						size_t n);
+#endif
+
+
+#ifdef L___wcslcpy
+#define L_strlcpy
+#define Wstrlcpy __wcslcpy
+#ifdef __LOCALE_C_ONLY
+weak_alias(__wcslcpy,wcsxfrm);
+#endif
+#endif
+
 #ifdef L_strlcpy
 
+#ifndef L___wcslcpy
+#define Wstrlcpy strlcpy
+#ifdef __LOCALE_C_ONLY
+weak_alias(strlcpy,strxfrm);
+#endif
+#endif
+
 /* OpenBSD function:
  * Copy at most n-1 chars from src to dst and nul-terminate dst.
  * Returns strlen(src), so truncation occurred if the return value is >= n. */
 
-size_t strlcpy(register char *__restrict dst,
-			   register const char *__restrict src,
-			   size_t n)
+size_t Wstrlcpy(register Wchar *__restrict dst,
+				  register const Wchar *__restrict src,
+				  size_t n)
 {
-	const char *src0 = src;
-	char dummy[1];
+	const Wchar *src0 = src;
+	Wchar dummy[1];
 
 	if (!n) {
 		dst = dummy;
@@ -1917,3 +2170,621 @@ void psignal(int signum, register const char *message)
 
 #endif
 /**********************************************************************/
+#ifndef __LOCALE_C_ONLY
+
+#ifdef L_strxfrm
+#ifndef WANT_WIDE
+#error WANT_WIDE should be defined for L_strxfrm
+#endif
+#ifdef L_wcsxfrm
+#error L_wcsxfrm already defined for L_strxfrm
+#endif
+
+#define wcscoll strcoll
+#define L_wcsxfrm
+#undef WANT_WIDE
+
+#undef Wvoid
+#undef Wchar
+#undef Wuchar
+#undef Wint
+
+#define Wchar char
+
+#endif /* L_strxfrm */
+
+
+
+#ifdef L_wcsxfrm
+
+#define CUR_COLLATE (&__global_locale.collate)
+
+#define MAX_PENDING 8
+
+typedef struct {
+	const Wchar *s;
+	const Wchar *eob;			/* end of backward */
+
+	__uwchar_t weight;
+	__uwchar_t ui_weight;		/* undefined or invalid */
+	int colitem;
+	int weightidx;
+	int rule;
+	size_t position;
+	/* should be wchar_t.  if wchar < 0 do EILSEQ? */
+	__uwchar_t *cip;
+	__uwchar_t ci_pending[MAX_PENDING];	/* nul-terminated */
+
+	char *back_buf;
+	char *bbe;					/* end of back_buf (actual last... not 1 past end) */
+	char *bp;					/* ptr into backbuf, NULL if not in backward mode */
+	char ibb[128];
+	size_t bb_size;
+
+	int ru_pushed;
+} col_state_t;
+
+
+#define WEIGHT_MASK	0x3fffU
+#define RULE_MASK	0xc000U
+
+#define RULE_FORWARD  (1 << 14)
+#define RULE_POSITION (1 << 15)
+
+#define UI_IDX		(WEIGHT_MASK-6)
+#define POSIT_IDX	(WEIGHT_MASK-5)
+#define RANGE_IDX	(WEIGHT_MASK-4)
+#define UNDEF_IDX	(WEIGHT_MASK-3)
+#define INVAL_IDX	(WEIGHT_MASK-2)
+#define DITTO_IDX   (WEIGHT_MASK-1)
+
+
+#undef TRACE
+#if 0
+#define TRACE(X)	printf##X
+#else
+#define TRACE(X)	((void)0)
+#endif
+
+static int lookup(wchar_t wc)
+{
+	unsigned int sc, n, i0, i1;
+
+	if (((__uwchar_t) wc) > 0xffffU) {
+		return 0;
+	}
+
+	sc = wc & CUR_COLLATE->ti_mask;
+	wc >>= CUR_COLLATE->ti_shift;
+	n = wc & CUR_COLLATE->ii_mask;
+	wc >>= CUR_COLLATE->ii_shift;
+
+	i0 = CUR_COLLATE->wcs2colidt_tbl[wc];
+	i0 <<= CUR_COLLATE->ii_shift;
+	i1 = CUR_COLLATE->wcs2colidt_tbl[CUR_COLLATE->ii_len + i0 + n];
+	i1 <<= CUR_COLLATE->ti_shift;
+	return CUR_COLLATE->wcs2colidt_tbl[CUR_COLLATE->ii_len + CUR_COLLATE->ti_len + i1 + sc];
+
+}
+
+static void init_col_state(col_state_t *cs, const Wchar *wcs)
+{
+	memset(cs, 0, sizeof(col_state_t));
+	cs->s = wcs;
+	cs->bp = cs->back_buf = cs->ibb;
+	cs->bb_size = 128;
+	cs->bbe = cs->back_buf + (cs->bb_size -1);
+}
+
+static void next_weight(col_state_t *cs, int pass)
+{
+	int r, w, ru, ri, popping_backup_stack;
+	ssize_t n;
+	const uint16_t *p;
+#ifdef WANT_WIDE
+#define WC (*cs->s)
+#define N (1)
+#else  /* WANT_WIDE */
+	mbstate_t mbstate;
+	wchar_t WC;
+	size_t n0, nx;
+#define N n0
+
+	mbstate.mask = 0;
+#endif /* WANT_WIDE */
+
+	do {
+
+		if (cs->ru_pushed) {
+			ru = cs->ru_pushed;
+			TRACE(("ru_pushed = %d\n", ru));
+			cs->ru_pushed = 0;
+			goto POSITION_SKIP;
+		}
+
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning should we walk pendings backwards?
+#endif
+		if (cs->cip) {			/* possible pending weight */
+			if ((r = *(cs->cip++)) == 0) {
+				cs->cip = NULL;
+				continue;
+			}
+			cs->weightidx = r & WEIGHT_MASK;
+			assert(cs->weightidx);
+/* 			assert(cs->weightidx != WEIGHT_MASK); */
+		} else {				/* get the next collation item from the string */
+			TRACE(("clearing popping flag\n"));
+			popping_backup_stack = 0;
+
+		IGNORE_LOOP:
+			/* keep first pos as 0 for a sentinal */
+			if (*cs->bp) {				/* pending backward chars */
+			POP_BACKUP:
+				popping_backup_stack = 1;
+				TRACE(("setting popping flag\n"));
+				n = 0;
+				if (*cs->bp > 0) {		/* singles pending */
+					cs->s -= 1;
+					if ((*cs->bp -= 1) == 0) {
+						cs->bp -= 1;
+					}
+				} else {				/* last was a multi */
+					cs->s += *cs->bp;
+					cs->bp -= 1;
+				}
+			} else if (!*cs->s) { /* not in backward mode and end of string */
+				cs->weight = 0;
+				return;
+			} else {
+				cs->position += 1;
+			}
+
+		BACK_LOOP:
+#ifdef WANT_WIDE
+			n = 1;
+			cs->colitem = r = lookup(*cs->s);
+#else  /* WANT_WIDE */
+			n = n0 = mbrtowc(&WC, cs->s, SIZE_MAX, &mbstate);
+			if (n < 0) {
+				__set_errno(EILSEQ);
+				cs->weight = 0;
+				return;
+			}
+			cs->colitem = r = lookup(WC);
+#endif /* WANT_WIDE */
+
+			TRACE((" r=%d WC=%#lx\n", r, (unsigned long)(WC)));
+
+			if (r > CUR_COLLATE->max_col_index) { /* starting char for one or more sequences */
+				p = CUR_COLLATE->multistart_tbl;
+				p += p[r-CUR_COLLATE->max_col_index -1];
+				do {
+					n = N;
+					r = *p++;
+					do {
+						if (!*p) {		/* found it */
+							cs->colitem = r;
+							TRACE(("    found multi %d\n", n));
+							goto FOUND;
+						}
+#ifdef WANT_WIDE
+						/* the lookup check here is safe since we're assured that *p is a valid colidx */
+						if (!cs->s[n] || (lookup(cs->s[n]) != *p)) {
+							do {} while (*p++);
+							break;
+						}
+						++p;
+						++n;
+#else  /* WANT_WIDE */
+						if (cs->s[n]) {
+							nx = mbrtowc(&WC, cs->s + n, SIZE_MAX, &mbstate);
+							if (nx < 0) {
+								__set_errno(EILSEQ);
+								cs->weight = 0;
+								return;
+							}
+						}
+						if (!cs->s[n] || (lookup(WC) != *p)) {
+							do {} while (*p++);
+							break;
+						}
+						++p;
+						n += nx;
+#endif /* WANT_WIDE */
+					} while (1);
+				} while (1);
+			} else if (r == 0) {		/* illegal, undefined, or part of a range */
+				if ((CUR_COLLATE->range_count)
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning .. need to introduce range as a collating item?
+#endif
+					&& (((__uwchar_t)(WC - CUR_COLLATE->range_low)) <= CUR_COLLATE->range_count)
+					) {					/* part of a range */
+					/* Note: cs->colitem = 0 already. */
+					TRACE(("    found range\n"));
+					ru = CUR_COLLATE->ruletable[CUR_COLLATE->range_rule_offset*CUR_COLLATE->MAX_WEIGHTS + pass];
+					assert((ru & WEIGHT_MASK) != DITTO_IDX);
+					if ((ru & WEIGHT_MASK) == WEIGHT_MASK) {
+						ru = (ru & RULE_MASK) | RANGE_IDX;
+						cs->weight = CUR_COLLATE->range_base_weight + (WC - CUR_COLLATE->range_low);
+					}
+					goto RANGE_SKIP_TO;
+				} else if (((__uwchar_t)(WC)) <= 0x7fffffffUL) { /* legal but undefined */
+				UNDEFINED:
+					/* Note: cs->colitem = 0 already. */
+					ri = CUR_COLLATE->undefined_idx;
+					assert(ri != 0); /* implicit undefined isn't supported */
+
+					TRACE(("    found explicit UNDEFINED\n"));
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning right now single weight locales do not support ..
+#endif
+					if (CUR_COLLATE->num_weights == 1) {
+						TRACE(("    single weight UNDEFINED\n"));
+						cs->weightidx = RANGE_IDX;
+						cs->weight = ri;
+						cs->s += n;
+						goto PROCESS_WEIGHT;
+					}
+
+					ri = CUR_COLLATE->index2ruleidx[ri - 1];
+					ru = CUR_COLLATE->ruletable[ri * CUR_COLLATE->MAX_WEIGHTS + pass];
+					assert((ru & WEIGHT_MASK) != WEIGHT_MASK); /* TODO: handle ".." */
+					if ((ru & WEIGHT_MASK) == DITTO_IDX) {
+						cs->colitem = CUR_COLLATE->undefined_idx;
+					}
+					goto RANGE_SKIP_TO;
+				} else {		/* illegal */
+					TRACE(("    found illegal\n"));
+					__set_errno(EINVAL);
+					/* We put all illegals in the same equiv class with maximal weight,
+					 * and ignore them after the first pass. */
+					if (pass > 0) {
+						cs->s += n;
+						goto IGNORE_LOOP;
+					}
+					ru = (RULE_FORWARD | RANGE_IDX);
+					cs->weight = 0xffffU;
+					goto RANGE_SKIP_TO;
+				}
+			} else if (CUR_COLLATE->num_weights == 1) {
+				TRACE(("    single weight\n"));
+				cs->weightidx = RANGE_IDX;
+				cs->weight = cs->colitem;
+				cs->s += n;
+				goto PROCESS_WEIGHT;
+			} else {
+				TRACE(("    normal\n"));
+			}
+
+			/* if we get here, it is a normal char either singlely weighted, undefined, or in a range */
+		FOUND:
+			ri = CUR_COLLATE->index2ruleidx[cs->colitem - 1];
+			TRACE((" ri=%d ", ri));
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning make sure this is correct
+#endif
+			if (!ri) {
+				TRACE(("NOT IN THIS LOCALE\n"));
+				goto UNDEFINED;
+			}
+			ru = CUR_COLLATE->ruletable[ri * CUR_COLLATE->MAX_WEIGHTS + pass];
+
+		RANGE_SKIP_TO:
+
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning ignoreables probably should not interrupt backwards processing, but this is wrong
+#endif
+/* 			if (!(ru & WEIGHT_MASK)) { */
+/* 				TRACE(("IGNORE\n")); */
+/* 				cs->s += n; */
+/* 				continue; */
+/* 			} */
+
+
+			TRACE((" rule = %#x  weight = %#x  popping = %d  s = %p  eob = %p\n",
+				   ru & RULE_MASK, ru & WEIGHT_MASK, popping_backup_stack,
+				   cs->s, cs->eob));
+			/* now we need to check if we're going backwards... */
+
+			if (!popping_backup_stack) {
+				if (!(ru & RULE_MASK)) { /* backward */
+					TRACE(("backwards\n"));
+					assert(cs->bp <= cs->bbe);
+					if (cs->bp == cs->bbe) {
+						if (cs->back_buf == cs->ibb) { /* was using internal buffer */
+							cs->bp = malloc(cs->bb_size + 128);
+							if (!cs->bp) {
+								__set_errno(ENOMEM);
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning what to do here?
+#endif
+								cs->weight = 0;
+								return;
+							}
+							memcpy(cs->bp, cs->back_buf, cs->bb_size);
+
+						} else {
+							cs->bp = realloc(cs->back_buf, cs->bb_size + 128);
+							if (!cs->bp) {
+								__set_errno(ENOMEM);
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning what to do here?
+#endif
+								cs->weight = 0;
+								return;
+							}
+						}
+						cs->bb_size += 128;
+						cs->bbe = cs->bp + (cs->bbe - cs->back_buf);
+						cs->back_buf = cs->bp;
+						cs->bp = cs->bbe;
+
+					}
+					if (n==1) {			/* single char */
+						if (*cs->bp && (((unsigned char)(*cs->bp)) < CHAR_MAX)) {
+							*cs->bp += 1; /* increment last single's count */
+						} else {	  /* last was a multi, or just starting */
+							if (!cs->bp) {
+								cs->bp = cs->back_buf;
+							} else {
+								assert(cs->bp < cs->bbe);
+								++cs->bp;
+							}
+							*cs->bp = 1;
+						}
+					} else {			/* multichar */
+						assert(n>1);
+						assert(cs->bp < cs->bbe);
+						*++cs->bp = -n;
+					}
+					cs->s += n;
+					if (*cs->s) {
+						goto BACK_LOOP;
+					}
+					/* end-of-string so start popping */
+					cs->eob = cs->s;
+					TRACE(("popping\n"));
+					goto POP_BACKUP;
+				} else if (*cs->bp) { /* was going backward but this element isn't */
+					/* discard current and use previous backward element */
+					assert(!cs->cip);
+					cs->eob = cs->s;
+					TRACE(("popping\n"));
+					goto POP_BACKUP;
+				} else {				/* was and still going forward */
+					TRACE(("forwards\n"));
+					if ((ru & (RULE_POSITION|WEIGHT_MASK)) > RULE_POSITION) {
+						assert(ru & WEIGHT_MASK);
+						cs->ru_pushed = ru;
+						cs->weight = cs->position;
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning devel code
+#endif
+						cs->position = 0;	/* reset to reduce size for strcoll? */
+						cs->s += n;
+						cs->weightidx = RANGE_IDX;
+						goto PROCESS_WEIGHT;
+					}
+				}
+			} else {					/* popping backwards stack */
+				TRACE(("popping (continued)\n"));
+				if (!*cs->bp) {
+					cs->s = cs->eob;
+				}
+				cs->s -= n;
+			}
+
+			cs->s += n;
+		POSITION_SKIP:
+			cs->weightidx = ru & WEIGHT_MASK;
+			cs->rule = ru & RULE_MASK;
+		}
+
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning for pending we only want the weight... _not_ the rule
+#endif
+		if (!cs->weightidx) {	/* ignore */
+			continue;
+		}
+
+	PROCESS_WEIGHT:
+		assert(cs->weightidx);
+
+
+		if (((unsigned int)(cs->weightidx - UI_IDX)) <= (INVAL_IDX-UI_IDX)) {
+			if (cs->weightidx == UI_IDX) {
+				cs->weight = cs->ui_weight;
+			}
+			return;
+		}
+
+		assert(cs->weightidx != WEIGHT_MASK);
+		if (cs->weightidx == DITTO_IDX) { /* want the weight of the current collating item */
+			TRACE(("doing ditto\n"));
+			w = CUR_COLLATE->index2weight[cs->colitem -1];
+		} else if (cs->weightidx <= CUR_COLLATE->max_col_index) { /* normal */
+			TRACE(("doing normal\n"));
+			w = CUR_COLLATE->index2weight[cs->weightidx -1];
+		} else {				/* a string */
+			TRACE(("doing string\n"));
+			assert(!(cs->weightidx & RULE_MASK));
+			/* note: iso14561 allows null string here */
+			p = CUR_COLLATE->weightstr + (cs->weightidx - (CUR_COLLATE->max_col_index + 2));
+			if (*p & WEIGHT_MASK) {
+				r = 0;
+				do {
+					assert(r < MAX_PENDING);
+					cs->ci_pending[r++] = *p++;
+				} while (*p & WEIGHT_MASK);
+				cs->cip = cs->ci_pending;
+			}
+			continue;
+		}
+
+		cs->weight = w;
+		return;
+	} while (1);
+}
+
+int wcscoll (const Wchar *s0, const Wchar *s1)
+{
+	col_state_t ws[2];
+	int pass;
+
+	if (!CUR_COLLATE->num_weights) { /* C locale */
+#ifdef WANT_WIDE
+		return wcscmp(s0, s1);
+#else  /* WANT_WIDE */
+		return strcmp(s0, s1);
+#endif /* WANT_WIDE */
+	}
+
+	pass = 0;
+	do {						/* loop through the weights levels */
+		init_col_state(ws, s0);
+		init_col_state(ws+1, s1);
+		do {					/* loop through the strings */
+			/* for each string, get the next weight */
+			next_weight(ws, pass);
+			next_weight(ws+1, pass);
+			TRACE(("w0=%lu  w1=%lu\n",
+				   (unsigned long) ws[0].weight,
+				   (unsigned long) ws[1].weight));
+
+			if (ws[0].weight != ws[1].weight) {
+				return ws[0].weight - ws[1].weight;
+			}
+		} while (ws[0].weight);
+	} while (++pass < CUR_COLLATE->num_weights);
+
+	return 0;
+}
+
+#ifdef WANT_WIDE
+
+size_t wcsxfrm(wchar_t *__restrict ws1, const wchar_t *__restrict ws2, size_t n)
+{
+	col_state_t cs;
+	size_t count;
+	int pass;
+
+	if (!CUR_COLLATE->num_weights) { /* C locale */
+		return __wcslcpy(ws1, ws2, n);
+	}
+
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning handle empty string as a special case
+#endif
+
+	count = pass = 0;
+	do {						/* loop through the weights levels */
+		init_col_state(&cs, ws2);
+		do {					/* loop through the string */
+			next_weight(&cs, pass);
+			TRACE(("weight=%lu (%#lx)\n", (unsigned long) cs.weight, (unsigned long) cs.weight));
+			if (count < n) {
+				ws1[count] = cs.weight +1;
+			}
+			++count;
+			TRACE(("--------------------------------------------\n"));
+		} while (cs.weight);
+		if (count <= n) {		/* overwrite the trailing 0 end-of-pass marker */
+			ws1[count-1] = 1;
+		}
+		TRACE(("--------------------  pass %d  --------------------\n", pass));
+	} while (++pass < CUR_COLLATE->num_weights);
+	if (count <= n) {			/* oops... change it back */
+		ws1[count-1] = 0;
+	}
+	return count-1;
+}
+
+#else  /* WANT_WIDE */
+
+static const unsigned long bound[] = {
+	1UL << 7,
+	1UL << 11,
+	1UL << 16,
+	1UL << 21,
+	1UL << 26,
+};
+
+static unsigned char first[] = {
+	0x0, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc
+};
+
+/* Use an extension of UTF-8 to store a 32 bit val in max 6 bytes. */
+
+static size_t store(unsigned char *s, size_t count, size_t n, __uwchar_t weight)
+{
+	int i, r;
+
+	i = 0;
+	do {
+		if (weight < bound[i]) {
+			break;
+		}
+	} while (++i < sizeof(bound)/sizeof(bound[0]));
+
+	r = i+1;
+	if (i + count < n) {
+		s += count;
+		s[0] = first[i];
+		while (i) {
+			s[i] = 0x80 | (weight & 0x3f);
+			weight >>= 6;
+			--i;
+		}
+		s[0] |= weight;
+	}
+
+	return r;
+}
+
+size_t strxfrm(char *__restrict ws1, const char *__restrict ws2, size_t n)
+{
+	col_state_t cs;
+	size_t count, inc;
+	int pass;
+
+	if (!CUR_COLLATE->num_weights) { /* C locale */
+		return strlcpy(ws1, ws2, n);
+	}
+
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning handle empty string as a special case
+#endif
+
+	inc = count = pass = 0;
+	do {						/* loop through the weights levels */
+		init_col_state(&cs, ws2);
+		do {					/* loop through the string */
+			next_weight(&cs, pass);
+			TRACE(("weight=%lu (%#lx)\n", (unsigned long) cs.weight, (unsigned long) cs.weight));
+			inc = store((unsigned char *)ws1, count, n, cs.weight + 1);
+			count += inc;
+			TRACE(("--------------------------------------------\n"));
+		} while (cs.weight);
+		/* overwrite the trailing 0 end-of-pass marker */
+		assert(inc == 1);
+		if (count <= n) {
+			ws1[count-1] = 1;
+		}
+		TRACE(("--------------------  pass %d  --------------------\n", pass));
+	} while (++pass < CUR_COLLATE->num_weights);
+	if (count <= n) {			/* oops... change it back */
+		ws1[count-1] = 0;
+	}
+	return count-1;
+}
+
+
+#endif /* WANT_WIDE */
+
+#endif /* wcscoll */
+
+#endif /* __LOCALE_C_ONLY */
+/**********************************************************************/
+
